#############################################
# reconstruct scores with specific residuals
#############################################
VECMB.ys <- function(x, b, p, e, type)
{
  n <- nrow(x); k <- nrow(b)
  b0 <- b[,1:(k*p),drop=FALSE]
  if(type=="const") b1 <- as.matrix(b[,(k*p)+1],nrow=k)
  if(type=="const+trend")
  {b1 <- as.matrix(b[,(k*p)+1],nrow=k); b2 <- as.matrix(b[,(k*p)+2],nrow=k)}
  y <- x[1:p, ,drop=FALSE]
  for( i in 1:(n-p) )
  {
    index <- 1:k
    d1 <- 0
    for(j in 1:p)
    {
      d <- b0[,index] %*% t(y[nrow(y)-j+1, , drop=FALSE])
      d1<- d1+d
      index<- index+k
    }
    d1 <- d1 + as.matrix(e[i,],nrow=k)
    if(type=="const") d1 <- d1 + b1
    if(type=="const+trend") d1 <- d1 + b1 + b2*i
    colnames(d1) <- NULL
    y <- rbind(y,t(d1))
  }
  return(y)
}


####################################
# FIT FVECM ON DEMOGAPHY DATA
####################################

fvecm <- function(demogdata, k)
{
  epsilon.f <- demogdata$smooth.err$female
  epsilon.m <- demogdata$smooth.err$male
  epsilon.f[is.na(epsilon.f)] <- 0
  epsilon.m[is.na(epsilon.m)] <- 0;
  smooth.var <- list(female = epsilon.f, male = epsilon.m)
  y.f <- log(demogdata$rate$female)
  y.m <- log(demogdata$rate$male)
  fts.obj.f <- fts(demogdata$age, y.f)
  fts.obj.m <- fts(demogdata$age, y.m)
  pca.f <- ftsm(fts.obj.f, order = k)
  pca.m <- ftsm(fts.obj.m, order = k)
  coef = vecm <- list()
  n <- length(demogdata$year)
  
  # fitted scores
  fitted.f = fitted.m <- array(NA, c(n-2, k))
  for(it in 1:k)
  {
    coef[[it]] <- cbind(pca.f$coeff[, it+1], pca.m$coeff[, it+1])
    vecm[[it]] <- VECM(coef[[it]], lag = 1, r = 1, include = "const")
    fitted <- fitted(vecm[[it]], level = "original")
    fitted.f[, it] <- fitted[, 1]
    fitted.m[, it] <- fitted[, 2]
  }
  
  # fitted curves
  fitted.curve.f <- pca.f$basis[,2:(k+1)]%*%t(fitted.f) + pca.f$basis[, 1]
  fitted.curve.m <- pca.m$basis[,2:(k+1)]%*%t(fitted.m) + pca.m$basis[, 1]
  fitted <- list(female = fitted.curve.f, male = fitted.curve.m)
  
  # residual curves
  resid.f <- fitted.curve.f - y.f[, c(-1,-2)]
  resid.m <- fitted.curve.m - y.m[, c(-1,-2)]
  residuals <- list(female = resid.f, male = resid.m)
  coeffs <- list(female = pca.f$coeff, male = pca.m$coeff)
  basis <- list(female = pca.f$basis, male = pca.m$basis)
  score <- list(coefficients = coef, model = vecm)
  return(list(dim = k, fitted = fitted, residuals = residuals, score = score,
              age = demogdata$age, year = demogdata$year, smooth.var = smooth.var, 
              coeffs = coeffs, basis = basis))
}

####################################
# FORECAST FVECM
####################################

forecast.fvecm <- function(object, h = 20, level = 0.8, nboot = 100)
{
  n <- length(object$year); m <- length(object$age)
  k <- object$dim
  score.f.boot <- score.m.boot <- array(NA, dim = c(k, nboot, h))
  boot.curve.f <- boot.curve.m <- array(NA, dim = c(m, h, nboot))
  fore.obj.f <- fore.obj.m <- array(NA, dim = c(age+1, h))
  pred.f <- pred.m <- array(NA, dim = c(h, k))
  for(it in 1:k)
  {
    pred <- predict(object$score$model[[it]], n.ahead = h)
    pred.f[, it] <- pred[,1]
    pred.m[, it] <- pred[,2]
  }
  fore.obj.f <- object$basis$female[, 2:(k+1)]%*%t(pred.f) + object$basis$female[, 1]
  fore.obj.m <- object$basis$male[, 2:(k+1)]%*%t(pred.m) + object$basis$male[, 1]
  
  for (b in 1:nboot)
  {
    for(it in 1:k)
    {
      index <- sample(n-2, replace = T)
      # sample the residual of it-th pair of scores
      res <- object$score$model[[it]]$resid
      res.sampled <- res[index,]
      A1 <- diag(1, 2) + object$score$model[[it]]$coefficients[, 3:4] + t(object$score$model[[it]]$model.specific$beta%*%object$score$model[[it]]$coefficients[, 1])
      A2 <- -object$score$model[[it]]$coefficients[, 3:4]
      B <- cbind(A1, A2, object$score$model[[it]]$coefficients[, 2])
      # reconstruct the it-th pair of scores with the sampled residuals
      reconstructed.score <- VECMB.ys(object$score$coefficients[[it]], b = B, p =2, res.sampled, type = "const")
      # refit vecm model with reconstructed scores
      vecm.boot <- VECM(reconstructed.score, lag = 1, r = 1, include = "const")
      A1 <- diag(1, 2) + vecm.boot$coefficients[, 3:4] + t(vecm.boot$model.specific$beta%*%vecm.boot$coefficients[, 1])
      A2 <- -vecm.boot$coefficients[, 3:4]
      B <- cbind(A1, A2, vecm.boot$coefficients[, 2])
      pred.boot <- VAR.etp:::VAR.ForeB(object$score$coefficients[[it]], b = B, p = 2, h = h, res)
      #make predictions using the bootstrapped coefficient and the original data
      score.f.boot[it,b,] <- pred.boot[, 1]
      score.m.boot[it,b,] <- pred.boot[, 2]
    }
    
    ind.f <- sample(n-h-1, h, replace = TRUE)
    ind.m <- sample(n-h-1, h, replace = TRUE)
    
    # construct nboot sets of bootstrapped curves
    boot.curve.f[,,b] <- object$basis$female[, 2:(k+1)]%*%score.f.boot[,b,] + object$basis$female[, 1] + object$residuals$female[, ind.f] +  matrix(rnorm(m*h, 0, rep(object$smooth.var$female[,ncol(object$smooth.var$female)], h)),m, h)
    boot.curve.m[,,b] <- object$basis$male[, 2:(k+1)]%*%score.m.boot[,b,] + object$basis$male[, 1] + object$residuals$male[, ind.m] +  matrix(rnorm(m*h, 0, rep(object$smooth.var$male[,ncol(object$smooth.var$male)], h)),m, h)
  }
  boot.sample <- list(female = boot.curve.f, male = boot.curve.m)
  Forecast <- list(female = fore.obj.f, male = fore.obj.m)
  pi.f <- VAR.etp:::VAR.PI(boot.curve.f, level, fore.obj.f)$PI
  pi.m <- VAR.etp:::VAR.PI(boot.curve.m, level, fore.obj.m)$PI
  PIs <- list(female = pi.f, male = pi.m)
  return(list(boot.sample = boot.sample, Forecast = Forecast, Intervals = PIs))
}

